# $Id: server.py 4075 2010-11-29 09:15:20Z place $
# $URL: https://svn.non-gnu.uvt.nl/uvt-dev/tags/sources/meresco/bzv-9.5.1/src/server.py $
import pprint
import gc

#FIXME: mag weg?
infra = None

from meresco.components.statistics import Statistics, Logger

from meresco.components import Lxml2Amara, XmlXPath, XmlPrintLxml, Amara2Lxml, Venturi, Configuration, FilterField, RenameField, TransformFieldValue, RewritePartname, Xml2Fields, XPath2Field, FilterMessages, Fields2XmlTx, Reindex
from meresco.components.statisticsxml import StatisticsXml
from meresco.components.drilldown import SRUTermDrilldown
from meresco.components.http import PathFilter, FileServer, PathRename
from meresco.components.sru import SruParser, SruHandler, SRURecordUpdate
from meresco.components.tokenizefieldlet import TokenizeFieldlet

from meresco.core import Transparant, TransactionScope,  BatchTransactionScope, ResourceManager, Observable

from meresco.components.facetindex.tools import unlock

from bzv.facetindex import Drilldown, CQL2LuceneQuery, Fields2LuceneDocumentTx, DrilldownFieldnames
from bzv.facetindex.merescolucene import IndexWriter, merescoStandardAnalyzer
from bzv.facetindex.lucene import LuceneIndex

# ImportError: No module named ngram
#from meresco.components.ngram import NGramFieldlet, NGramQuery, RatioSuggester

from os.path import join, dirname, abspath, isdir
from os import remove, makedirs, unlink
from glob import iglob

# unicode normalization
from unicodedata import normalize as uNormalize
normalForm = "NFKC"

## BZV specifics ###

# Anton: use homebrew dynamichtml from bzv instead of the cq2 version to support logging
from bzv.dynamichtml import DynamicHtml

from bzv.delfttilburg.normalize import LastnameNormalize, FirstInitialNormalize
from bzv.delfttilburg import RateRecordCookie, RateRecord, MultiLevelDrilldown, ComposeNameFromMods, ModsDateField, DidYouMean

from bzv.drilldownconfig import DrilldownConfig
from bzv.yearnormalize import YearNormalize
from bzv.solr import Solr
from bzv.rss import Rss
from bzv.rssitem import RssItem
from bzv.utils import HandleRequestPrinter, LimitQuery, Printer, NormalizeQuery
from bzv.openurl import OpenURL
from bzv.composedaifrommods import ComposeDaiFromMods
from bzv.composedatefrommods import ComposeDateFromMods
from bzv.composedatestampfromheader import ComposeDateStampFromHeader
from bzv.composefrommods import ComposeFromMods
from bzv.composefulljournaltitlefrommods import ComposeFullJournalTitleFromMods
from bzv.composejelfrommods import ComposeJELFromMods
from bzv.composesetsfromheader import ComposeSetsFromHeader
from bzv.record import Record, RecordXml, Citec, RecordJSON
from bzv.json import Request, Response
from bzv.xmldocumentadapter import XmlDocumentAdapter
from bzv.writedeleted import WriteDeleted
from bzv.undelete import Undelete
from bzv.getavailability import GetAvailability
from bzv.aselect import Aselect
from bzv.storagecomponent import BZVStorageComponent
from bzv.getidentifiers import GetIdentifiers
from bzv.observablewebquery import ObservableWebQuery
from bzv.access import RecordAccess
from bzv.metadata2mods import Metadata2Mods
from bzv.modsidentifierfield import ModsIdentifierField
from bzv.addtimestamp import AddTimestamp

def optimize(databasePath):
    weggegooid = False
    indexPath = join(databasePath, 'index')
    for fs in ['write.lock', 'tracker.segments', '*.docids', '*.deleted']:
        for f in iglob(join(indexPath, fs)):
            unlink(f)
            weggegooid = True

    if not weggegooid:
        return

    unlock(indexPath)
    writer = IndexWriter(indexPath, merescoStandardAnalyzer, False)
    writer.optimize()
    writer.close()

class PathMangler(Observable):
    # prefix en param zouden een lijstje kunnen worden
    def __init__(self, prefix, param, **kwargs):
        Observable.__init__(self)
        self._prefix = prefix
        self._param = param
        self._kwargs = kwargs

    def handleRequest(self, **kwargs):
        mangled = kwargs
        path = kwargs['path']
        if path.startswith(self._prefix + '/'):
            mangled['path'] = self._prefix
            mangled['arguments'][self._param] = [path[len(self._prefix) + 1:]]
            mangled['arguments'].update(self._kwargs)
        return self.all.handleRequest(**mangled)

class SpyingEye(Observable):
    def __init__(self, infra):
        Observable.__init__(self)
        self.infra = infra
        self.logger = infra.logger
        self.logger.info("Spying eye")
        self.pp = pprint.PrettyPrinter(indent=2)

    def getInfra(self):
        return self.infra

    def spy(self):
        return self.logger.debug

    def spit(self,text):
        self.logger.debug("Spit: >>>>>>>>>>>>>>>>\n" + self.pp.pformat(text) + "\n<<<<<<<<<<<<<<<<"  )


class GarbageCollector(Observable):
    def handleRequest(self, Client=None, *args, **kwargs):
        gc.collect()
        return self.all.handleRequest(Client=Client, *args, **kwargs)

class ServerLog(Observable, Logger):
    def handleRequest(self, Client=None, *args, **kwargs):
        if Client:
            self.log(ip=Client[0])
        self.log(request='request')
        return self.all.handleRequest(Client=Client, *args, **kwargs)

class OurLogPrinter(Observable):
    def __init__(self, name):
        Observable.__init__(self)
        self._name = name

    def unknown(self, *args, **kwargs):
        print self._name, args, kwargs
        return self.all.unknown(*args, **kwargs)

class StatusReduce(Observable):
    def addField(self, name, value):
        if len(value) == 0:
            value = 'done'
        else:
            value = reduce(lambda a,b: a if a == 'done' else b, value)
        if value == '':
            value = 'failed'
        return self.do.addField(name=name, value=value)

def createUploadHelix(index, storageComponent, xsltFiles, reactor, infra, nsMap):
    yearNormalize = YearNormalize(yearRange=(1000,2100))
    firstInitialNormalize = FirstInitialNormalize()
    lastnameNormalize = LastnameNormalize()
    logger=infra.logger

    configRoot =  infra.configRoot
    timeout = int(configRoot.get('timeout',section='batchTransactions',default="1"))
    batchSize = int(configRoot.get('batchSize',section='batchTransactions',default="50"))

    DRILLDOWN_PREFIX = configRoot.get('DRILLDOWN_PREFIX', section = 'drilldown')
    SORT_PREFIX = configRoot.get('SORT_PREFIX', section = 'drilldown')

    untokenized = []
    untokenizedDict = configRoot.getSectionDictList('untokenizedFields')

    for key in untokenizedDict:
        untokenized.extend(untokenizedDict[key])

    fields2LuceneDocument = \
        (ResourceManager("record",
                         lambda resourceManager:
                             Fields2LuceneDocumentTx(resourceManager,
                                                     untokenized=untokenized)),
         index
    )

    indexingHelix = \
        (Transparant(),
            fields2LuceneDocument,
            (FilterField(lambda name: DRILLDOWN_PREFIX + name in  untokenizedDict['drilldownfieldnames']),
                (RenameField(lambda name: DRILLDOWN_PREFIX + name),
                    fields2LuceneDocument
                )
            ),
            (FilterField(lambda name: SORT_PREFIX + name in untokenizedDict['sortfieldnames']),
                (RenameField(lambda name: SORT_PREFIX + name),
                    fields2LuceneDocument
                )
            ),
            (TransformFieldValue(lambda value: uNormalize(normalForm, unicode(value))),
                    fields2LuceneDocument
            )
        )

    normalizeTransactionStoreHelix = \
        (ResourceManager("record", lambda resourceManager:Fields2XmlTx(resourceManager, 'extra', namespace="http://meresco.com/namespace/fields/extra")),
            (storageComponent,)
        )

    storeModsHelix = \
        (RewritePartname('mods'),
            (XmlPrintLxml(),
                (storageComponent,)
            ),
        )

    indexModsHelix = \
        (Transparant(),    
            (Xml2Fields(),
                # aparte mods-velden indexeren
                indexingHelix,
                (RenameField(lambda name: name.split('.')[0]),
                    # veld mods indexeren
                    indexingHelix
                ),
                (FilterField(lambda name: name == '__id__'),
                    normalizeTransactionStoreHelix
                )
            ),
            (ComposeNameFromMods('norm.name.oneinitial'),
                (TransformFieldValue(firstInitialNormalize.normalize),
                    indexingHelix,
                    normalizeTransactionStoreHelix,
                ),
                (TransformFieldValue(firstInitialNormalize.unparsable),
                    (RenameField(lambda name: 'norm.name.unparsable'),
                        indexingHelix
                    )
                ),
                (TransformFieldValue(lastnameNormalize.normalize),
                    (RenameField(lambda name: 'norm.name.lastname'),
                        indexingHelix,
                        normalizeTransactionStoreHelix,
                    )
                )
            ),
            (ComposeFromMods(infra, 'norm.genre'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeFromMods(infra, 'norm.language'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeDaiFromMods('norm.dai'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeFullJournalTitleFromMods('norm.fulljournaltitle'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeJELFromMods('norm.jel'),
                indexingHelix,
                normalizeTransactionStoreHelix,
            ),
            (ComposeDateFromMods('norm.date.year'),
                (TransformFieldValue(yearNormalize.normalize),
                    indexingHelix,
                    normalizeTransactionStoreHelix,
                    (TransformFieldValue(lambda year: year[:len('YYY')]+'0'),
                        (RenameField(lambda name: 'norm.date.decennium'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    (TransformFieldValue(lambda year: year[:len('YY')]+'00'),
                        (RenameField(lambda name: 'norm.date.centennium'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    )
                ),
                (TransformFieldValue(yearNormalize.unparsable),
                    (RenameField(lambda name: 'norm.date.unparsable'),
                        indexingHelix
                    )
                )
            ),
            (ModsDateField(),
                indexingHelix,
            ),
            (ModsIdentifierField(),
                indexingHelix,
            ),
        )

    indexDDI3Helix = \
        (XmlXPath(['/ddi3:DDIInstance/s:StudyUnit'], namespaceMap=nsMap),
            (XPath2Field([
                ('//c:ConceptualComponent/c:GeographicStructureScheme/r:GeographicStructure/r:Geography/r:Level/r:Name/text()', 'ddi.geography'),
                ('//r:Coverage/r:TopicalCoverage/r:Keyword/text()', 'ddi.keyword'),
                ('//c:ConceptualComponent/c:ConceptScheme/c:Concept/r:Label/text()', 'ddi.concept'),
                ('//l:LogicalProduct/l:VariableScheme/l:Variable/r:Label/text()', 'ddi.variable')
                ], namespaceMap=nsMap),
                indexingHelix,
                normalizeTransactionStoreHelix
            ),
        )

    return \
        (BatchTransactionScope("batch", reactor=reactor,
                               batchSize=batchSize,
                               timeout=timeout),
            (TransactionScope("record"),
                (Venturi(
                        should=[
                            ('metadata', '/document:document/document:part[@name="metadata"]/text()'),
                            ('meta', '/document:document/document:part[@name="meta"]/text()')],
                        could=[
                            ('rating',   '/rating:rating'),
                            ('header',   '/document:document/document:part[@name="header"]/text()'),
                            ('fulltext', '/fulltext:fulltext'),
                            ('ddi',      '/document:document/document:part[@name="ddi"]/text()'),
                            ('ddi3',     '/document:document/document:part[@name="ddi3"]/text()'),
                            ('jel',      '/document:document/document:part[@name="jel"]/text()')],
                        namespaceMap = nsMap
                    ),
                    (FilterMessages(allowed=['delete']),
                        index,
                        (WriteDeleted(),
                        (storageComponent,))
                    ),
                    (FilterMessages(allowed=['getStream']),
                        (storageComponent,)
                    ),

################### TP: storing of the required ('should') and optional ('could') parts
                    (XmlPrintLxml(),
                         (storageComponent,)
                    ),
                    (XmlXPath(['/oai:metadata'], namespaceMap=nsMap),
                        (Metadata2Mods(xsltFiles, nsMap),
                            indexModsHelix,
                        )
                    ),
                    (XPath2Field([("//oai:metadata/didl:DIDL/didl:Item/didl:Descriptor/didl:Statement/dii:Identifier/text()", 'norm.docid')], namespaceMap=nsMap),
                        indexingHelix,
                        normalizeTransactionStoreHelix,
                    ),
                    ### fulltext indexing ###
                    (XPath2Field([("//didl:Component/didl:Resource[@mimeType='application/pdf']/@ref", 'fulltext.url.exists')], namespaceMap=nsMap),
                        (TransformFieldValue(lambda value: 'true'),
                            indexingHelix
                        )
                    ),
                    (XmlXPath(['/fulltext:fulltext'], namespaceMap=nsMap),
                        (RewritePartname('fulltext'),
                            (XmlPrintLxml(),
                                 (storageComponent,)
                            ),
                        ),
                        (Xml2Fields(),
                            indexingHelix
                        ),
                        (XPath2Field([('//fulltext:status/text()', 'fulltext.item.status')], namespaceMap=nsMap, sendAsList=True),
                            (StatusReduce(),
                                indexingHelix
                            ),
                        ),
                    ),
                    ## set from header
                    (XmlXPath(['/oai:header'], namespaceMap=nsMap),
                        #(Printer('header-set'),),
                        (ComposeSetsFromHeader('set'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    ## datestamp form header
                    (XmlXPath(['/oai:header'], namespaceMap=nsMap),
                        #(Printer('header-datestamp'),),
                        (ComposeDateStampFromHeader('datestamp'),
                            indexingHelix,
                            normalizeTransactionStoreHelix
                        )
                    ),
                    (XmlXPath(['/meta:meta'], namespaceMap=nsMap),
                        (Xml2Fields(),
                            indexingHelix,
                            (FilterField(lambda name: name == 'meta.repository.id'),
                                (RenameField(lambda name: 'meresco.exists'),
                                    (TransformFieldValue(lambda value: 'true'),
                                        indexingHelix
                                    )
                                )
                            ),
                            (RenameField(lambda name: 'meta'),
                                indexingHelix
                            )
                        ),
                        (AddTimestamp(),
                            (storageComponent,),
                            indexingHelix
                        )
                    ),
                    (XmlXPath(['/ddi3:DDIInstance'], namespaceMap=nsMap),
                        indexDDI3Helix
                    ),
                    (FilterMessages(disallowed=['delete']),
                        (Undelete(),
                        (storageComponent,))
                    )
                )
            )
        )

def main(reactor, observableHttpServer, sruUpdateServer, inbox, baseURL, portNumber, databasePath, infra, configWatcher):
    logger = infra.logging.getLogger('meresco.server')

    configRoot =  infra.configRoot
    configDict = configRoot.getSectionDict('system')
    appid = configRoot.get('aselectid', section = 'authentication')

    drilldownFieldnames = configRoot.getValueAsList ('drilldownFieldnames', section='untokenizedFields')
    DRILLDOWN_PREFIX = configRoot.get('DRILLDOWN_PREFIX', section = 'drilldown')
    nsMap = configRoot.getSectionDict('namespaces')
    
#    logger.debug(infra.prpt('nsMap', nsMap))
    unqualifiedTermFields = configRoot.getSectionTupleList('unqualifiedTermFields')

    applicationName = configRoot.get('applicationName',section='userFeedback', default='Beter Zoeken en Vinden')
    storageDirectory = join(databasePath, 'storage')
    indexDirectory = join(databasePath, 'index')

    dynamicPath                = configDict['dynamicpath']
    staticPath                 = configDict['staticpath']
    xsltPath                   = configDict['xsltpath']
    templatePath               = configDict['templatepath']
    logPath                    = configDict['logpath']
    reindexIdentifierBatchPath = configDict['reindexidentifierbatchpath']
    reindexPartName            = configDict['reindexpartname']
    startWithoutFacetIndex     = configDict['startwithoutfacetindex']

    luceneDebugLogFilename = configRoot.get('name', section='lucenedebuglogfile')
    if luceneDebugLogFilename: luceneDebugLogFilename = join(logPath, luceneDebugLogFilename)

    # space seperated filelist
    xsltFileList = configDict['xsltfiles'].split()
    xsltFiles = [join(xsltPath, f) for f in xsltFileList ]

    unlock(indexDirectory)

    storageType = configRoot.get('storagetype', section = 'storagecomponent', default = 'boushi')
    storageComponent = BZVStorageComponent(storageDirectory, partsRemovedOnDelete=[], nsMap=nsMap, xslt=xsltFiles, storageType=storageType)
    logger.debug('storageType: ' + storageType)

    drilldownComponent = Drilldown(drilldownFieldnames, transactionName="batch")
    if luceneDebugLogFilename:
        luceneIndex = LuceneIndex(indexDirectory, transactionName="batch", infra = infra, debugLogFilename = luceneDebugLogFilename )
    else:
        luceneIndex = LuceneIndex(indexDirectory, transactionName="batch", infra = infra)
    if startWithoutFacetIndex == 'yes':
        print "Starting without facet index"
        indexHelix = \
            (luceneIndex,
            )
    else: 
        indexHelix = \
            (luceneIndex,
                (drilldownComponent,)
            )
    uploadHelix = createUploadHelix(indexHelix, storageComponent, xsltFiles, reactor=reactor,infra=infra, nsMap=nsMap)

    configurationComponent = Configuration(configDict)
    rateRecordList = RateRecordCookie(join(databasePath, 'ratedRecords'))

    # statistics
    statisticsPath = join(databasePath, 'statistics')
    isdir(statisticsPath) or makedirs(statisticsPath)
    #statistics = Statistics(statisticsPath, [('ip',), ('clause',)], 60)
    statistics = Statistics(statisticsPath, [('clause',)], 60)

    locator = configRoot.get('Locator',section='proxies')
    if configRoot.get('Testing', section='locator') == 'True':
        print "Testing locator"
        locator = "/availability" + configRoot.get('Locator',section='proxies')
    else:
        print "Using acceptation/production locator"

    recordAccess = (RecordAccess(infra),)

    record = \
        (Record(infra),
            (storageComponent,),
            (CQL2LuceneQuery(unqualifiedTermFields),
                indexHelix),
            (Record(infra),
                (storageComponent,)
            )
        )

    sru = \
        (SruParser(host=baseURL, port=portNumber, defaultRecordSchema='metadata', defaultRecordPacking='xml'),
            (SruHandler(),
                (CQL2LuceneQuery(unqualifiedTermFields),
                    indexHelix
                ),
                (storageComponent,),
                (SRUTermDrilldown(True),
                    (DrilldownFieldnames(
                        lambda field: DRILLDOWN_PREFIX + field),
                            (drilldownComponent,)
                    ),
                    (CQL2LuceneQuery(unqualifiedTermFields),
                        indexHelix
                    ),
                )
            )
        )

    multiLevelDrilldown = \
        (MultiLevelDrilldown(infra),
            (SpyingEye(infra),),
            (DrilldownConfig(infra),),
            (drilldownComponent,),
        )

    return \
        (Observable(),
            (configWatcher,),
            (inbox,
                (XmlDocumentAdapter(),
                    uploadHelix,
                )
            ),
            (sruUpdateServer,
                (PathFilter('/update'),
                    (SRURecordUpdate(),
                        (Amara2Lxml(),
                            uploadHelix
                        )
                    )
                ),
                (PathFilter('/reindex'),
                    (Reindex(partName=reindexPartName, filelistPath=reindexIdentifierBatchPath),
                        (FilterMessages(allowed=['listIdentifiers']),
                            (storageComponent,)
                        ),
                        (FilterMessages(disallowed=['listIdentifiers']),
                            uploadHelix
                        )
                    )
                )
            ),
            (observableHttpServer,
                (statistics,
                    (Aselect(infra, '^', baseURL, appid),
                        (SpyingEye(infra),),
                        (ServerLog(), ),
                        (PathFilter('/statistics'),
                            (StatisticsXml(statistics), )
                        ),
                        (NormalizeQuery(normalForm),),
                        (PathFilter('/rss'),
                            #FIXME use config!
                            (Rss(   title = applicationName + " the most recent results for the query: %(query)s (click a title to get it) ",
                                    description = 'This service is brought to you by Tilburg University Library and IT Services, lis-servicedesk@uvt.nl',
                                    link = baseURL,
                                    sortKeys = 'datestamp,,1',
                                    maximumRecords = 15),
                                (CQL2LuceneQuery(unqualifiedTermFields),
                                    indexHelix
                                ),
                                (RssItem(
                                    nsMap=nsMap,
                                    pubDate =     ('extra', '/extra:extra/extra:datestamp/text()'),
                                    linkTemplate = baseURL + '/id/%(uploadId)s',
                                    uploadId =    ('meta', '/meta:meta/meta:upload/meta:id/text()')),
                                    (storageComponent,),
                                    recordAccess,
                                    record
                                ),
                            )
                        ),
                        (PathFilter('/record'),
                            (RecordXml(),
                                recordAccess,
                                record
                            )
                        ),
                        (PathFilter('/citec'),
                            (Citec(),
                                recordAccess,
                                record
                            )
                        ),
                        (PathFilter('/JSON/record'),
                            (RecordJSON(),
                                recordAccess,
                                record
                            )
                        ),
                        (PathFilter('/JSON/searchresult'),
                            (HandleRequestPrinter(),),
                            (Request(infra),
                                (ObservableWebQuery(), ),
                                (CQL2LuceneQuery(unqualifiedTermFields),
                                    indexHelix
                                ),
                                (Record(infra),
                                    (storageComponent,)
                                ),
                                (drilldownComponent,),
                                multiLevelDrilldown,

                                (SpyingEye(infra),),
                                (DrilldownConfig(infra),),
                            )
                        ),
####################### TP: this is only for testing
                        (PathFilter('/availability'),
                            (GetAvailability(infra),)
                        ),
                        (PathFilter('/getidentifiers'),
                            #(HandleRequestPrinter(),),
                            (GetIdentifiers(infra),
                                (CQL2LuceneQuery(unqualifiedTermFields),
                                    indexHelix
                                )
                            )
                        ),
                        (PathFilter('/eo/sru'),
                            (HandleRequestPrinter(),),
                            (LimitQuery(" AND (meta.repository.collection exact neeo OR meta.repository.collection exact repec)"),
                                sru
                            )
                        ),
                        (PathFilter('/sru'),
                            (HandleRequestPrinter(),),
                            sru
                        ),
                        (PathFilter('/files'),\
                            (PathRename(lambda path: path[len('/files'):]),
                                (FileServer(staticPath),)
                            )
                        ),
                        (PathFilter('/gc'),\
                            (GarbageCollector(),)
                        ),
                        (PathFilter(['/', '/m/search', '/search', '/id'] , excluding=[ '/rss', '/record', '/citec', '/JSON/record', '/JSON/searchresult', '/availability', '/getidentifiers', '/eo/sru', '/sru', '/update', '/files', '/statistics', '/reindex', '/gc']),
                            (PathMangler('/id', 'query', displayType=['single']),
                                (DynamicHtml([dynamicPath], reactor=reactor, allowedModules=['cqlparser', 'cqlwebutils'],
                                             indexPage='/search',
                                             infra = infra,
                                             templatePath = templatePath,
                                             baseURL = baseURL
                                             ),
                                    (SpyingEye(infra),),
                                    (ObservableWebQuery(), ),
                                    (CQL2LuceneQuery(unqualifiedTermFields),
                                        #(Solr("http://acceptatie.solr.uvt.nl/solr/select/"),),
                                        indexHelix
                                    ),
                                    (Response(infra), 
                                        (DrilldownConfig(infra),),
                                    ),
                                    (OpenURL(locator, infra),
                                        recordAccess,
                                        record    
                                    ),
                                    recordAccess,
                                    record,
                                    (storageComponent,),
                                    (drilldownComponent,),
                                    (RateRecord(1,5),
                                        (FilterMessages(allowed=['isAvailable', 'getStream']),
                                            (storageComponent,),
                                        ),
                                        uploadHelix
                                    ),
                                    multiLevelDrilldown,
                                    (configurationComponent,),
                                    (rateRecordList, ),
                                    (DrilldownConfig(infra),),
                                )
                            )
                        )
                    )
                )
            )
        )
